#include <iomanip>
#include <vector>
#include <algorithm>
#include <iostream>
#include <fstream>
#include <string>
#include <sstream>
#include <stdio.h>
#include <cmath>
#include <ctime>
#include <cstdlib>
#include <sys/stat.h> 
#include <sys/time.h> 
#include <gsl/gsl_sf.h> 
#include <gsl/gsl_interp.h>
#include <gsl/gsl_spline.h>
#include <getopt.h> 

#include "LHAPDF/LHAPDF.h"  // interface with LHAPDF

#include "complex.hh"
#include "mellin_Dk.cc"
#include "NLO_finite-mt/bdv.hh"
#include "NNLOpl.hh"
#include "NNLO_finite-mt/nnlo_finite-mt.hh"
#include "ggHiggs.hh"

#include <cuba.h>


using namespace std;



const double ZETA2  =  1.64493406684822643647;
const double ZETA3  =  1.2020569031595942855;
const double ZETA4  =  1.0823232337111381915;
const double ZETA5  =  1.03692775514336992633136548646;
const double CA = 3.;
const double CF = 4./3.;
const int _nf = 5;


// PRIVATE flags and variables ( don't change here !!! )
double mtop = 172.5;
double mbot = 4.4; // negative mass means no bottom running in the loop
double mcha = 1.2; // negative mass means no charm running in the loop
int    order = 1;
//int PDForder = 1;
int NSET = 1;
bool finite_mt = false;
SmallxCoeffs *sxc;
double _muF_mu_ratio  = 1.;
double _muR_muF_ratio = 1.;
bool __full__ = false;
bool __soft_part__ = true;
bool __hard_part__ = true;
int __log_type__ = 0; // 0:D, 1:Dhat, 2:Dlog
bool __Nsoftcoll__ = false;
bool __NNLL__ = false;
int APtype = 0;
bool APimprExponent = true;
bool MSbar = true;
bool quiet = true;
double INTEGRAL_PRECISION = 15.e-3;







double Li2(double x){
 double x_0 = -0.30;
 double x_1 = 0.25;
 double x_2 = 0.51;
 if (x == 1.) return ZETA2;
 if (x <= x_0){ 
   double temp = log(ABS(1.0-x));
   return -Li2(-x/(1.0-x)) - temp*temp/2 ; }
 else if (x < x_1){
   double z = - log(1.0-x);
   double temp = z*(1.0-z/4.0*(1.0-z/9.0*(1.0-z*z/100.0
                  *(1.0-5.0*z*z/294.0*(1.0-7.0*z*z/360.0
                  *(1.0-5.0*z*z/242.0*(1.0-7601.0*z*z/354900.0
                  *(1.0-91.0*z*z/4146.0*(1.0-3617.0*z*z/161840.0)
                   ))))))));
   return temp; }
   else if (x < x_2) return - Li2(-x) + Li2(x*x)/2.0 ;
   else { return ZETA2 - Li2(1.0-x) 
                  - log(ABS(x))*log(ABS(1.0-x)) ; }
}
double Li3(double x){
 double x_0 = -1.0;
 double x_1 = -0.85;
 double x_2 = 0.25;
 double x_3 = 0.63;
 double x_4 =  1.0;
 if (x == 1.) return ZETA3;
 if (x == -1.) return - 0.75 * ZETA3;
 if (x <= x_0){ 
   double lnx = log(-x);
   return Li3(1.0/x) - ZETA2*lnx - lnx*lnx*lnx/6.0; }
 else if (x < x_1){
   return Li3(x*x)/4.0 - Li3(-x); }
   else if (x < x_2){
     double z = - log(1.0-x);
     double temp = z*(1.0-3.0*z/8.0*(1.0-17.0*z/81.0*(1.0-15*z/136.0
                    *(1.0-28.0*z/1875.0*(1.0+5.0*z/8.0*(1.0-304.0*z/7203.0
                    *(1.0+945.0*z/2432.0*(1.0-44.0*z/675.0*(1.0+7.0*z/24.0
                    *(1.0-26104.0*z/307461.0*(1.0+1925.0*z/8023.0
                    *(1.0-53598548.0*z/524808375.0
                    *(1.0+22232925.0*z/107197096.0
                     )))))))))))));
     return temp; }
     else if (x < x_3){
       return Li3(x*x)/4.0 - Li3(-x); }
       else if (x < x_4){
         double ln1x = log(1.0-x); 
         return -Li3(1.0-x) - Li3(-x/(1.0-x)) + ZETA3 + ZETA2*ln1x
	   - log(x)*ln1x*ln1x/2.0 + ln1x*ln1x*ln1x/6.0; }
       else { 
         double lnx = log(x);
         return Li3(1./x) + 2.0*ZETA2*lnx - lnx*lnx*lnx/6.0; }
}






// 1-d integration
#include<gsl/gsl_integration.h>
double gauss(double f(double,void*), double xmin, double xmax, double prec, double *error, void *par) {
  double result, err;
  gsl_function F;
  gsl_integration_workspace * w = gsl_integration_workspace_alloc (1000);
  F.function = f;
  F.params=par;
  gsl_integration_qags (&F, xmin, xmax, 0, prec, 1000,    w, &result, &err);
  gsl_integration_workspace_free (w);
  if(error) *error = err;
  return result;
}


// alpha_s running from ihixs
double alphas_ihixs(double mu, int porder, double as0, double mu0=mZ) {
  // Nf = 5 **************** <---------
  double bb0 = 1.916666667;
  double bb1 = 1.260869565;
  double bb2 = 1.474788647;
  double bb3 = 9.835916431;
  //
  double as;
  as0 /= M_PI;
  //double as0 = asZ/M_PI;
  //double as0=asZ/3.1415926;
  //double Mz=91.187;
  //
  int N = 100000;
  //double step = 2.*log(mu/mZ)/N, incr;
  //double step = 2.*log(mu/Mz)/N, incr;
  double step = 2.*log(mu/mu0)/N, incr;
  for (int i=0; i<N; i++) {
    if (porder >= 3)
      incr = step*bb0*as0*as0*(1+bb1*as0 + bb2*as0*as0 + bb3*as0*as0*as0 );
    else if (porder == 2)
      incr = step*bb0*as0*as0*(1+bb1*as0 + bb2*as0*as0 );
    else if (porder == 1)
      incr = step*bb0*as0*as0*(1+bb1*as0);
    else if (porder == 0)
      incr = step*bb0*as0*as0;
    as = as0 - incr;
    as0 = as;
  }
  return as*M_PI;
}







double b0 = (33.-2.*_nf)/12.;
double b1 = (153.-19.*_nf)/24.;
double b2 = (2857. - 5033./9.*_nf + 325./27.*_nf*_nf)/128.;


dcomplex InverseMellinIntegrand(dcomplex N, double log1ox, dcomplex f(dcomplex, void*), void *par) {
  return exp( N*log1ox ) * f(N,par) / M_PI;
}
double InverseMellin(double z, dcomplex f(dcomplex, void*), void *par) { // Talbot algorithm
  if(z==1.) return 0;
  double log1ox = -log(z);
  double res = 0.; 
  // prec ~ 10^{-0.6*M}
  int M = 18;
  double r = 1.+2.*M/5./log1ox;
  double t, cott, sigma;
  dcomplex N;
  res = real(0.5 * InverseMellinIntegrand(r, log1ox, f, par));
  for(int k=1; k<M; k++) {
    t = k*M_PI/M;
    cott = 1./tan(t);
    sigma = t + cott*(t*cott - 1.);
    N = r*t*(cott+I);
    res += real(InverseMellinIntegrand(N, log1ox, f, par) * (1.+I*sigma));
  }
  res *= M_PI*r/M;
  return res;
}




dcomplex gamma0poles(dcomplex N, int nf) {
  double e01 = CA;
  double e00 = (-11.*CA+2.*nf*(2.*CF/CA-1.))/12.;
  return (e01/(N-1) + e00);
}
dcomplex gamma1poles(dcomplex N, int nf) {
  double e11 = nf*(26.*CF-23.*CA)/36.;
  return e11/(N-1);
}
dcomplex gamma2poles(dcomplex N, int nf) {
  double e22 = CA*( (18.*ZETA2-71.)*(CA-2*CF)*nf + CA*CA*(54*ZETA3+99.*ZETA2-395) ) /108.;
  return e22/(N-1)/(N-1);
}
dcomplex FO_1_sx_NLL_RC(dcomplex N, double mu) {
  double lf = -2.*log(_muF_mu_ratio);
  double c10 = sxc->c10(mu,mtop);
  double c1 = 2*c10;
  c1 += 2.*lf;
  return c1*gamma0poles(N,_nf);
}
dcomplex FO_2_sx_NLL_RC(dcomplex N, double mu) {
  double lf = -2.*log(_muF_mu_ratio);
  double c10 = sxc->c10(mu,mtop);
  double c20 = sxc->c20(mu,mtop);
  double c11 = sxc->c11(mu,mtop);
  double c1 = 2*c10;
  double c2 = 2.*c20+c11;
  c20 += (c1+lf)*lf/2.;
  c2  += (c1+lf)*lf*2.;
  c1  += 2*lf;
  dcomplex g0 = gamma0poles(N,_nf);
  dcomplex g1 = gamma1poles(N,_nf);
  return c2*g0*g0 - 2*c20*b0*g0 + c1*g1 ;
}
dcomplex FO_3_sx_NLL_RC(dcomplex N, double mu) {
  double lf = -2.*log(_muF_mu_ratio);
  double c10 = sxc->c10(mu,mtop);
  double c20 = sxc->c20(mu,mtop);
  double c11 = sxc->c11(mu,mtop);
  double c30 = sxc->c30(mu,mtop);
  double c21 = sxc->c21(mu,mtop);
  if(MSbar) c30 += 8./3.*ZETA3;
  double c1 = 2*c10;
  double c2 = 2.*c20+c11;
  double c3 = 2.*(c30+c21);
  // scale dependence
  c30 += c20*lf + c1*lf*lf/4. + lf*lf*lf/6.;
  c21 += (c2-c20)*lf + 3.*c1*lf*lf/4. + lf*lf*lf/2.;
  c20 += (c1+lf)*lf/2.;
  c2  += (c1+lf)*lf*2.;
  c1  += 2*lf;
  dcomplex g0 = gamma0poles(N,_nf);
  dcomplex g1 = gamma1poles(N,_nf);
  dcomplex g2 = gamma2poles(N,_nf);
  dcomplex hard = (c3*g0*g0*g0 - (3*c30+c21)*2*b0*g0*g0 + 4*c30*b0*b0*g0
		  + 2*c2*g0*g1 - 4*c20*b0*g1 + c1*g2 );
  //double lrf = 2.*log(_muR_muF_ratio);
  //hard += 4*b0*FO_2_sx_NLL_RC(N,mu)*lrf;
  return hard;
}
double MomConsCoeff[3] = {0,0,0};
void ComputeMomConsCoeff(double mu) {
  MomConsCoeff[0] = real(FO_1_sx_NLL_RC(2,mu) - 2.*FO_1_sx_NLL_RC(3,mu) + FO_1_sx_NLL_RC(4,mu));
  MomConsCoeff[1] = real(FO_2_sx_NLL_RC(2,mu) - 2.*FO_2_sx_NLL_RC(3,mu) + FO_2_sx_NLL_RC(4,mu));
  MomConsCoeff[2] = real(FO_3_sx_NLL_RC(2,mu) - 2.*FO_3_sx_NLL_RC(3,mu) + FO_3_sx_NLL_RC(4,mu));
}
void ResetMomConsCoeff() {
  MomConsCoeff[0] = 0;  MomConsCoeff[1] = 0;  MomConsCoeff[2] = 0;
}




double Dk(int k, double z) {
  return pow(log(1-z),k)/(1.-z);
}
double Dtk(int k, double z) {
  return pow(log(1-z)-0.5*log(z),k)/(1.-z);
}
double DMPk(int k, double z) {
  return pow(log(log(1./z)),k)/log(1./z);
}
double mell_const(int k, double fake=0) {
  return DG(k+1,1.)/(k+1.);
}




// perturbative expansions of constants and logs
double _cd1, _cd2, _cd3;
void init_delta_const() {
  _cd1 = 2.*CA * ZETA2;
  _cd2 = ( 837./16. + 67./2.*ZETA2 - 9./20.*ZETA2*ZETA2 - 165./4.*ZETA3
	   + _nf  * ( 5./6.*ZETA3 - 5./3.*ZETA2 - 247./36. ) );
  _cd3 = 1740.02 - 131.928*_nf + 1.75719*_nf*_nf; // new coefficient
}
double _W1, _W2, _W3;
double _W1t, _W2t, _W3t;
void init_Wilson(double mu) {
  double lHt = 2.*log(mu/mtop);
  _W1 = 11./2.;
  _W2 = 121./16. + 2777./144. + 19./8.*lHt + (-67./48. + 2./3.*lHt)*_nf;
  //_W3 = 40.6488 + 40.1458*lHt + 10.941*lHt*lHt; // write full expression .... done.
  _W3 = (-1792967./20736. +  897943./4608.*ZETA3 + 5347./288.*lHt + 209./32.*lHt*lHt
	 + _nf * (493./10368. + 209./54.*lHt + 23./16.*lHt*lHt - 110779./6912.*ZETA3)
	 + _nf*_nf * (-6865./15552. + 77./864.*lHt - lHt*lHt/9. ) );
}
void init_Wilson_mt(double mu) {
  _W1t = G2l();
  _W2t = gg_nnlo_mt_delta(mu, mtop, 1) - _cd2 - _cd1*_W1t;
  _W3t = _W3; // pointlike value
  if(mu > 2*mtop) _W2t = _W2; // pointlike value
}
const double A1 = CA*CA*(67./36.-ZETA2/2.) - 5./18.*CA*_nf;
const double A2 = CA*CA*CA*(245./24.-67./9.*ZETA2+11./6.*ZETA3+11./5.*ZETA2*ZETA2)/4.
  + CA*CA*_nf*(-209./108.+10./9.*ZETA2-7./3.*ZETA3)/4.
  + CA*CF*_nf*(-55./24.+2.*ZETA3)/4.
  - CA*_nf*_nf/4./27.;
const double B1 = CA*CA*(2./3.+3./4.*ZETA3) - (CA/6.+CF/8.)*_nf;
const double B2 = CA*CA*CA/4.*(79./32.-ZETA2*ZETA3+ZETA2/6.+11./24.*ZETA2*ZETA2+67./6.*ZETA3-5.*ZETA5)
  - CA*CA*_nf/4.*(233./288.+ZETA2/6.+ZETA2*ZETA2/12.+5./3.*ZETA3)
  - CA*CF*_nf/4.*241./288.
  + CF*CF*_nf/4./16.
  + _nf*_nf/4.*(CA*29./288.+CF*11./144.);
//
void init_consts(double mH) {
  initBDV(mH, mtop, mbot, mcha);
  init_delta_const();
  init_Wilson(mH);
  init_Wilson_mt(mH);
}



// Pointlike log contributions
double FO_1_log_PL(double z, double D(int,double)) {
  return 4.*CA*D(1,z);
}
double FO_2_log_PL(double z, double D(int,double)) {
  double value;
  value = ( 72. * D(3,z)
	    + (2.*_nf - 33.) * D(2,z)
	    + (67. - 90.*ZETA2 - 10./3.*_nf) * D(1,z)
	    + (-101./3. + 33.*ZETA2 + 351./2.*ZETA3 + _nf*(14./9.-2.*ZETA2)) * D(0,z)
	    );
  return value;
}
double FO_3_log_PL(double z, double D(int,double)) {
  double CA3 = CA*CA*CA;
  double value;
  value = ( 8*CA3 * D(5,z)
	    + (-110./9.*CA3 + 20/9.*CA*CA*_nf) * D(4,z)
	    + (CA3*(925./27. - 56*ZETA2) - 164./27.*CA*CA*_nf + 4./27.*CA*_nf*_nf) * D(3,z)
	    + (CA3 * (-1051./27. + 187./3*ZETA2 + 181*ZETA3)
	       + CA*CA*_nf * (457./2./27. - 34./3.*ZETA2)
	       + 0.5*CA*CF*_nf - 10./27.*CA*_nf*_nf ) * D(2,z)
	    + (CA3 * (30569./8./81. - 152./3.*ZETA2 - 352./3.*ZETA3 - 154./5.*ZETA2*ZETA2)
	       + CA*CA*_nf * (-4211./4./81. + 94./9.*ZETA2 + 46./3.*ZETA3)
	       - CA*CF*_nf * (63./8.-6.*ZETA3)
	       + CA*_nf*_nf * (25./81.-4./9.*ZETA2) ) * D(1,z)
	    + (CA3 * (-297029./32./729. + 8563./4./81.*ZETA2 + 8941./4./27.*ZETA3 + 253./4./15.*ZETA2*ZETA2
		      - 725./2./3.*ZETA2*ZETA3 + 186.*ZETA5)
	       + CA*CA*_nf * (31313./16./729. - 2173./4./81.*ZETA2 - 475./4./9.*ZETA3 - 17./2./15*ZETA2*ZETA2)
	       + CA*CF*_nf * (1711./32./27. - 0.5*ZETA2 - 19./2./9.*ZETA3 - 1./5.*ZETA2*ZETA2)
	       - CA*_nf*_nf * (58./729. - 10./27.*ZETA2 - 5./27.*ZETA3) ) * D(0,z)
	    );
  return value;
}
//
dcomplex mellD(int k, dcomplex N) {
  if(__log_type__ == 0)
    return mell_D(k,N);
  else if(__log_type__ == 1)
    return mell_DB(k,N);
  else if(__log_type__ == 2)
    return mell_DM(k,N);
  //return mell_DMconst(k,N);
  return 0;
}
//
dcomplex S1(dcomplex N) {
  return 4.*CA*mellD(1,N);
}
dcomplex S2(dcomplex N) {
  dcomplex res = ( (2./3.*_nf - 11./3.*CA)*CA * mellD(2,N)
		  + (CA*CA*(67./9. - 2.*ZETA2) - CA*10./9.*_nf) * mellD(1,N)
		  + (CA*CA*(-101./27. + 11./3.*ZETA2 + 7./2.*ZETA3) + _nf*CA*(14./27.-2./3.*ZETA2)) * mellD(0,N)
		  );
  return res;
}
dcomplex S1_sq(dcomplex N) {
  return 16*CA*CA*(mellD(3,N) - 2.*ZETA2*mellD(1,N) + 2.*ZETA3*mellD(0,N) - ZETA2*ZETA2/10.);
}
dcomplex S1_cube(dcomplex N) {
  return 64*CA*CA*CA*( 3./4.*mellD(5,N) - 6.*ZETA2*mellD(3,N) + 15.*ZETA3*mellD(2,N) -3./2.*ZETA2*ZETA2*mellD(1,N)
		       +(18.*ZETA5-12.*ZETA2*ZETA3)*mellD(0,N) + 5*ZETA3*ZETA3 -29./35.*ZETA2*ZETA2*ZETA2);
}
dcomplex S1_S2(dcomplex N) {
  double b22 = (2./3.*_nf - 11./3.*CA)*CA;
  double b21 = CA*CA*(67./9. - 2.*ZETA2) - CA*10./9.*_nf;
  double b20 = CA*CA*(-101./27. + 11./3.*ZETA2 + 7./2.*ZETA3) + _nf*CA*(14./27.-2./3.*ZETA2);
  return 4*CA*( 5./6.*b22*mellD(4,N) + b21*mellD(3,N)
		+ (3./2.*b20-3.*ZETA2*b22) *mellD(2,N)
		+ (6.*ZETA3*b22-2.*ZETA2*b21) *mellD(1,N)
		+ (-ZETA2*ZETA2*b22+2.*ZETA3*b21-ZETA2*b20) *mellD(0,N)
		+ (4.*ZETA5-2.*ZETA2*ZETA3)*b22 -1./10.*ZETA2*ZETA2*b21 +ZETA3*b20
		);
}
//
dcomplex FO_2_log_PL(dcomplex N) {
  dcomplex value;
  value = ( 72. * mellD(3,N)
	    + (2.*_nf - 33.) * mellD(2,N)
	    + (67. - 90.*ZETA2 - 10./3.*_nf) * mellD(1,N)
	    + (-101./3. + 33.*ZETA2 + 351./2.*ZETA3 + _nf*(14./9.-2.*ZETA2)) * mellD(0,N)
	    );
  return value;
}
dcomplex FO_3_log_PL(dcomplex N) {
  double CA3 = CA*CA*CA;
  dcomplex value;
  value = ( 8*CA3 * mellD(5,N)
	    + (-110./9.*CA3 + 20/9.*CA*CA*_nf) * mellD(4,N)
	    + (CA3*(925./27. - 56*ZETA2) - 164./27.*CA*CA*_nf + 4./27.*CA*_nf*_nf) * mellD(3,N)
	    + (CA3 * (-1051./27. + 187./3*ZETA2 + 181*ZETA3)
	       + CA*CA*_nf * (457./2./27. - 34./3.*ZETA2)
	       + 0.5*CA*CF*_nf - 10./27.*CA*_nf*_nf ) * mellD(2,N)
	    + (CA3 * (30569./8./81. - 152./3.*ZETA2 - 352./3.*ZETA3 - 154./5.*ZETA2*ZETA2)
	       + CA*CA*_nf * (-4211./4./81. + 94./9.*ZETA2 + 46./3.*ZETA3)
	       - CA*CF*_nf * (63./8.-6.*ZETA3)
	       + CA*_nf*_nf * (25./81.-4./9.*ZETA2) ) * mellD(1,N)
	    + (CA3 * (-297029./32./729. + 8563./4./81.*ZETA2 + 8941./4./27.*ZETA3 + 253./4./15.*ZETA2*ZETA2
		      - 725./2./3.*ZETA2*ZETA3 + 186.*ZETA5)
	       + CA*CA*_nf * (31313./16./729. - 2173./4./81.*ZETA2 - 475./4./9.*ZETA3 - 17./2./15*ZETA2*ZETA2)
	       + CA*CF*_nf * (1711./32./27. - 0.5*ZETA2 - 19./2./9.*ZETA3 - 1./5.*ZETA2*ZETA2)
	       - CA*_nf*_nf * (58./729. - 10./27.*ZETA2 - 5./27.*ZETA3) ) * mellD(0,N)
	    );
  return value;
}
//
dcomplex APimprovement(dcomplex f(dcomplex), dcomplex N) {
  dcomplex res = 0;
  if(APtype==0)  res = f(N);                                           // Ag(1)
  if(APtype==1)  res = f(N+1);                                         // Ag expanded to first order
  if(APtype==2)  res = 2*f(N  ) -3*f(N+1) +2*f(N+2);                   // full Ag expanded to second order
  if(APtype==-1) res =   f(N-1) -2*f(N  ) +3*f(N+1) -2*f(N+2) +f(N+3); // full Ag
  if(APtype==-2) res =(3*f(N+1) -2*f(N+2) +  f(N+3))/2.;               // soft Ags
  if(APtype==-3) res =   f(N-1);                                       // small-x
  return res;
}
double APimprovement(double z) {
  double res = 0;
  if(APtype==0)  res = 1;                                       // Ag(1)
  if(APtype==1)  res = z;                                       // Ag expanded to first order
  if(APtype==2)  res = (2-3*z+2*z*z);                           // full Ag expanded to second order
  if(APtype==-1) res = (1./z-2+3*z-2*z*z+z*z*z);                // full Ag
  if(APtype==-2) res = (3*z-2*z*z+z*z*z/2);                     // soft Ags
  if(APtype==-3) res = 1./z;                                    // small-x
  return res;
}







// Fixed-order (NLO)
///////////////////////////////////////
double FO_1_log(double z, double D(int,double)) {
  double lf = -2.*log(_muF_mu_ratio);
  return FO_1_log_PL(z,D) + 2*CA*D(0,z)*lf;
}
double FO_1_log_APimproved(double z, double D(int,double)) {
  return APimprovement(z)*FO_1_log(z,D);
}
double FO_1_Nsoftcoll_impr(double z) {
  return 2.*CA*(gsl_sf_psi(1.)-log(log(1./z)));
}
double FO_1_delta_const() {
  double W1 = (finite_mt ? _W1t : _W1);
  double c = _cd1 + W1;
  //double lrf = 2.*log(_muR_muF_ratio);
  //c += 2*b0*lrf;
  return c;
}
// finite mt R_gg term
double Rgg_zspace(double z, double mu) {
  double z0 = (mu<200 ? 0.0002 : 0.01);
  double c1 = sxc->c1marzani(mu,mtop);
  //double c1 = z0*Rgg(z0, mu*mu/z0) - 2.*log(z0);
  double rgg = (z<z0 ? (c1+2.*log(z))/z : Rgg(z,mu*mu/z));
  return rgg;
}
double Rqg_zspace(double z, double mu) {
  double z0 = 0.000002;  // apparently it behaves better than Rgg,
  double c1 = sxc->c1marzani(mu,mtop); // so the asymptotic is not really needed
  double rqg = (z<z0 ? (c1/2.*CF + CF*log(z))/z : Rqg(z,mu*mu/z));
  return rqg;
}
double Rqqbar_zspace(double z, double mu) {
  // This has no numerical instabilities
  return Rqqbar(z,mu*mu/z);
}
double FO_1_nonlog(double z, double mu) {
  double lf = -2.*log(_muF_mu_ratio);
  double smallx = (finite_mt ? CA*Rgg_zspace(z, mu) : - 11./2.*pow(1.-z,3.)/z );
  double nonlog = CA * 2.*( (1./z-2.+z*(1.-z))*(2.*log(1.-z)-log(z)) -log(z)/(1.-z) ) + smallx;
  return nonlog + 2.*CA*(1./z-2.+z-z*z)*lf;
}
double FO_1_hard(double z, double mu) {
  double lf = -2.*log(_muF_mu_ratio);
  double c1 = sxc->c1marzani(mu,mtop);
  c1 += 2.*lf;
  double e01 = CA;
  //double e00 = (-11.*CA+2.*_nf*(2.*CF/CA-1.))/12.; // delta term in z-space, cancelled by damping
  double hard = c1*e01/z;// + c1*e00;
  return (hard - 12.*MomConsCoeff[0]) * pow(1-z, 2);
}
double FO_1_qg(double z, double mu) {
  double lf = -2.*log(_muF_mu_ratio);
  double Pgq = CF*(1.+pow(1.-z,2))/z/2.;
  double smallx = (finite_mt ? Rqg_zspace(z, mu) : 2./3.*z - pow(1.-z,2)/z );
  return Pgq*(2.*log(1-z)-log(z)+lf) + smallx;
}
double FO_1_qqb(double z, double mu) {
  return (finite_mt ? Rqqbar_zspace(z, mu) : 32./27.*pow(1-z,3)/z );
}
//




// Fixed-order (NNLO)
///////////////////////////////////////
dcomplex FO_2_log(dcomplex N) {
  double W1 = (finite_mt ? _W1t : _W1);
  return FO_2_log_PL(N) + W1*S1(N);
}
double FO_2_log(double z, double D(int,double)) {
  double lf = -2.*log(_muF_mu_ratio);
  double W1 = (finite_mt ? _W1t : _W1);
  double Clog = FO_2_log_PL(z,D) + W1*FO_1_log_PL(z,D);
  Clog += ( 108*lf * D(2,z)
	    + (36.*lf*lf - (33-2*_nf)*lf) * D(1,z)
	    + (-(33-2*_nf)/4*lf*lf + (2.*CA*W1 + 67./2.-5./3.*_nf-5.*CA*CA*ZETA2)*lf) * D(0,z) );
  //double lrf = 2.*log(_muR_muF_ratio);
  //Clog += 3*b0*FO_1_log(z,D)*lrf;
  return Clog;
  //if(__full__ && finite_mt) return gg_nnlo_mt_distr(z, mu, mtop, _muF_mu_ratio); // no longer needed
}
dcomplex FO_2_log_APimprovement(dcomplex N, void *par) {
  return 0.5*(pow(APimprovement(S1,N),2) - APimprovement(S1_sq,N));
}
double FO_2_log_APimprovement_soft1(double z) {
  double piece = (Li3(z) - ZETA3  + 0.5*log(z)*(-ZETA2 + 1./6.*pow(log(z),2) - Li2(z) - 0.5*log(z)*log(1-z)))/(1.-z);
  return 8*CA*CA*z*piece;
}
double FO_2_log_APimprovement_soft2(double z) {
  double piece = (Li3(z) - ZETA3  + 0.5*log(z)*(-ZETA2 + 1./6.*pow(log(z),2) - Li2(z) - 0.5*log(z)*log(1-z)))/(1.-z);
  return 8*CA*CA*(-4.+4.*z+4.*ZETA2-12.*z*ZETA2
		  +8.*log(1.-z) + 2.*(-1.+ZETA2+z*(3.+ZETA2))*log(z) +2.*z*pow(log(z),2)
		  -1./12.*(1.+z)*pow(log(z),3) - 4.*log(1-z)* (ZETA2+z*(2.+ZETA2)+2.*(-1+z)*log(z))
		  +2.*pow(log(1-z),2)*(-4.+4.*z+log(z)+z*log(z)) + 2.*(1.+z)*(2.+2.*log(1-z)-log(z))*Li2(z)
		  +4.*(1+z)*Li3(1-z) + 2.*(1+z)*Li3(z) - 2.*(1+z)*ZETA3
		  + (2.-3.*z+2.*z*z)*piece);
}
double FO_2_log_APimproved(double z, double D(int,double)) {
  double res = APimprovement(z)*FO_2_log(z,D);
  //if(APimprExponent) res += InverseMellin(z,FO_2_log_APimprovement,NULL);
  if(APimprExponent) {
    if     (APtype == 1 && __log_type__ == 1) res += FO_2_log_APimprovement_soft1(z);
    else if(APtype == 2 && __log_type__ == 1) res += FO_2_log_APimprovement_soft2(z);
    else res += InverseMellin(z,FO_2_log_APimprovement,NULL);
  }
  return res;
}
dcomplex FO_2_Nsoftcoll_impr(dcomplex N, void *par) {
  return 2.*CA*log(N)/N*S1(N);
}
double FO_2_Nsoftcoll_impr(double z) {
  return InverseMellin(z,FO_2_Nsoftcoll_impr,NULL);
}
double FO_2_delta_const(double mu) {
  double lf = -2.*log(_muF_mu_ratio);
  double W1 = (finite_mt ? _W1t : _W1);
  double W2 = (finite_mt ? _W2t : _W2);
  double c = _cd2 + W2 + _cd1*W1;
  c += -18.*ZETA2*lf*lf + (-3./4. + 171./2.*ZETA3 + 1./4.*_nf - (33.-2.*_nf)/2.*ZETA2 - W1*(33.-2.*_nf)/12.)*lf;
  //if(finite_mt) c = gg_nnlo_mt_delta(mu, mtop, _muF_mu_ratio); // no longer needed (checked)
  //
  //double lrf = 2.*log(_muR_muF_ratio);
  //double muRF = _muR_muF_ratio;  _muR_muF_ratio = 1.;
  //c += (2*b1+3*b0*FO_1_delta_const())*lrf + 3*b0*b0*lrf*lrf;
  //_muR_muF_ratio = muRF;
  return c;
}
// sigma(z) = z C(z)
// C(z) = [D(z)]_+ + nonlog(z)
// sigma(z) = [D(z)]_+ - (1-z)D(z) + z*nonlog(z)
//          = [D(z)]_+ + reg(z)
// nonlog(z) = [ reg(z) + (1-z)D(z) ] /z
double FO_2_nonlog(double z, double mu) {
  double nonlog;
  if(!finite_mt) nonlog = ( C2reg_pointlike(z, _nf, _muF_mu_ratio) + (1.-z)*FO_2_log(z,Dk) ) /z;
  else {
    double lf = -2.*log(_muF_mu_ratio);
    double c1 = sxc->c1marzani(mu,mtop);
    double c2 = sxc->c2marzani(mu,mtop);
    c2  += (c1+lf)*lf*2.;
    if(mu<2*mtop)
      nonlog = ( gg_nnlo_mt_reg(z, mu, mtop, _muF_mu_ratio) + smallx_match_z(z, c2) + (1.-z)*gg_nnlo_mt_distr(z, mu, mtop, _muF_mu_ratio) ) /z;
    else {
      //nonlog = ( gg_nnlo_infmt_reg(z, mu, mtop, _muF_mu_ratio) + smallx_match_z(z, c2) + (1.-z)*gg_nnlo_infmt_distr(z, mu, mtop, _muF_mu_ratio) ) /z;
      nonlog = ( gg_nnlo_infmt_reg(z, mu, mtop, _muF_mu_ratio) + (1.-z)*gg_nnlo_infmt_distr(z, mu, mtop, _muF_mu_ratio) ) /z;
      //finite_mt = false;  nonlog = ( C2reg_pointlike(z, _nf, _muF_mu_ratio) + (1.-z)*FO_2_log(z,Dk) ) /z;  finite_mt = true;
      double nonlog1 = CA * 2.*( (1./z-2.+z*(1.-z))*(2.*log(1.-z)-log(z)) -log(z)/(1.-z) );
      nonlog += (_W1t-_W1)*nonlog1;
    }
  }
  //double lrf = 2.*log(_muR_muF_ratio);
  //nonlog += 3*b0*FO_1_nonlog(z,mu)*lrf;
  return nonlog;
}
//
double FO_2_hard(double z, double mu) {
  //double hard = SF(z, mu, FO_2_sx_NLL_RC); // ok checked
  //double lrf = 2.*log(_muR_muF_ratio);
  //hard += 3*b0*FO_1_hard(z,mu)*lrf;
  // z-space version: faster
  double lf = -2.*log(_muF_mu_ratio);
  double c10 = sxc->c10(mu,mtop);
  double c20 = sxc->c20(mu,mtop);
  double c11 = sxc->c11(mu,mtop);
  double c1 = 2*c10;
  double c2 = 2.*c20+c11;
  c20 += (c1+lf)*lf/2.;
  c2  += (c1+lf)*lf*2.;
  c1  += 2*lf;
  double e01 = CA;
  double e00 = (-11.*CA+2.*_nf*(2.*CF/CA-1.))/12.;
  double e11 = _nf*(26.*CF-23.*CA)/36.;
  double hard = -c2*e01*e01*log(z)/z + (2*c2*e00*e01-2*c20*b0*e01+c1*e11)/z;
  //
  return (hard - 12.*MomConsCoeff[1]) * pow(1-z, 2);
}
double FO_2_qg(double z, double mu) {
  double res;
  if(!finite_mt) res = C2reg_pointlike_qg(z, _nf, _muF_mu_ratio) /z;
  else {
    double lf = -2.*log(_muF_mu_ratio);
    double c10 = sxc->c10(mu,mtop);
    double c20 = sxc->c20(mu,mtop);
    double c11 = sxc->c11(mu,mtop);
    double c1 = 2*c10;
    double c2 = 2.*c20+c11;
    c2  += (c1+lf)*lf*2.;
    c20 += (c1+lf)*lf/2.;
    if(mu<2*mtop)
      res = ( qg_nnlo_mt(z, mu, mtop, _muF_mu_ratio) + smallx_match_z(z, (c2-c20)*CF/CA) ) /z;
    else {
      res = qg_nnlo_infmt(z, mu, mtop, _muF_mu_ratio) /z;
      double Pgq = CF*(1.+pow(1.-z,2))/z/2.;
      double res1 = Pgq*(2.*log(1-z)-log(z));
      res += (_W1t-_W1)*res1;
    }
  }
  return res;
}
double FO_2_qqb(double z, double mu) {
  double res;
  if(!finite_mt) res = C2reg_pointlike_qqb(z, _nf, _muF_mu_ratio) /z;
  else {
    double lf = -2.*log(_muF_mu_ratio);
    double c10 = sxc->c10(mu,mtop);
    double c20 = sxc->c20(mu,mtop);
    double c11 = sxc->c11(mu,mtop);
    double c1 = 2*c10;
    double c2 = 2.*c20+c11;
    c2  += (c1+lf)*lf*2.;
    c20 += (c1+lf)*lf/2.;
    if(mu<2*mtop)
      res = ( qu_nnlo_mt(z, mu, mtop, _muF_mu_ratio) + smallx_match_z(z, (c2-2*c20)*pow(CF/CA,2)) ) /z;
    else {
      res = qu_nnlo_infmt(z, mu, mtop, _muF_mu_ratio) /z;
    }
  }
  return res;
}
double FO_2_qq(double z, double mu) {
  double res;
  if(!finite_mt) res = C2reg_pointlike_qq(z, _nf, _muF_mu_ratio) /z;
  else {
    double lf = -2.*log(_muF_mu_ratio);
    double c10 = sxc->c10(mu,mtop);
    double c20 = sxc->c20(mu,mtop);
    double c11 = sxc->c11(mu,mtop);
    double c1 = 2*c10;
    double c2 = 2.*c20+c11;
    c2  += (c1+lf)*lf*2.;
    c20 += (c1+lf)*lf/2.;
    if(mu<2*mtop)
      res = ( qq_nnlo_mt(z, mu, mtop, _muF_mu_ratio) + smallx_match_z(z, (c2-2*c20)*pow(CF/CA,2)) ) /z;
    else {
      res = qq_nnlo_infmt(z, mu, mtop, _muF_mu_ratio) /z;
    }
  }
  return res;
}
double FO_2_qqp(double z, double mu) {
  double res;
  if(!finite_mt) res = C2reg_pointlike_qqp(z, _nf, _muF_mu_ratio) /z;
  else {
    double lf = -2.*log(_muF_mu_ratio);
    double c10 = sxc->c10(mu,mtop);
    double c20 = sxc->c20(mu,mtop);
    double c11 = sxc->c11(mu,mtop);
    double c1 = 2*c10;
    double c2 = 2.*c20+c11;
    c2  += (c1+lf)*lf*2.;
    c20 += (c1+lf)*lf/2.;
    if(mu<2*mtop)
      res = ( qqp_nnlo_mt(z, mu, mtop, _muF_mu_ratio) + smallx_match_z(z, (c2-2*c20)*pow(CF/CA,2)) ) /z;
    else {
      res = qqp_nnlo_infmt(z, mu, mtop, _muF_mu_ratio) /z;
    }
  }
  return res;
}




// Fixed-order (nNNLO)
///////////////////////////////////////
dcomplex FO_3_log(dcomplex N) {
  double W1 = (finite_mt ? _W1t : _W1);
  double W2 = (finite_mt ? _W2t : _W2);
  return FO_3_log_PL(N) + W1*FO_2_log_PL(N) + W2*S1(N);
}
double D0D1(double z, double D(int,double)) {
  return 3./2.*D(2,z) - ZETA2*D(0,z);
}
double D0D2(double z, double D(int,double)) {
  return 4./3.*D(3,z) - 2.*ZETA2*D(1,z) + 2.*ZETA3*D(0,z);
}
double D0D3(double z, double D(int,double)) {
  return 5./4.*D(4,z) - 3.*ZETA2*D(2,z) + 6.*ZETA3*D(1,z) - 6.*ZETA4*D(0,z);
}
double FO_3_log(double z, double D(int,double)) {
  double lf = -2.*log(_muF_mu_ratio);
  double W1 = (finite_mt ? _W1t : _W1 );
  double W2 = (finite_mt ? _W2t : _W2 );
  double Clog = FO_3_log_PL(z,D) + W1*FO_2_log_PL(z,D) + W2*FO_1_log_PL(z,D);
  if(__NNLL__) {
    Clog -= 0.953938*D(0,z) *M_PI*M_PI*M_PI;
  }
  Clog += ( 0*lf * D(4,z)
	    + (16.*CA*CA*CA*lf*lf + 0*lf) * D(3,z)
	    + (4.*CA*CA*CA*lf*lf*lf - 18.*CA*CA*b0*lf*lf + 12.*CA*A1*lf) * D(2,z)
	    + (-4.*CA*CA*b0*lf*lf*lf + (8.*CA*A1+4.*CA*b0*b0+4.*CA*CA*W1-32.*CA*CA*CA*ZETA2)*lf*lf + (8*CA*B1-12*CA*b1)*lf) * D(1,z)
	    + ((2.*b0*b0-4.*CA*CA*CA*ZETA2)*lf*lf*lf + (CA*(4.*B1-5.*b1)-3.*CA*b0*W1-2.*A1*b0+6.*CA*CA*b0*ZETA2+32.*CA*CA*CA*ZETA3)*lf*lf
	       + (2.*A2+2.*W1*A1-4.*CA*A1*ZETA2)*lf) * D(0,z) );
  // adding contributions from interference with C2
  Clog += lf*(2.*CA*( 72. * D0D3(z,D)
		      + (2.*_nf - 33.) * D0D2(z,D)
		      + (67.+4*CA*W1 - 90.*ZETA2 - 10./3.*_nf) * D0D1(z,D)
		      + (-101./3. + 33.*ZETA2 + 351./2.*ZETA3 + _nf*(14./9.-2.*ZETA2)) * 2*D(1,z)
		      + (_cd2 + W2 + _cd1*W1 ) * D(0,z)
		      )
	      -2.*b0*( 72. * D(3,z)
		       + (2.*_nf - 33.) * D(2,z)
		       + (67.+4*CA*W1 - 90.*ZETA2 - 10./3.*_nf) * D(1,z)
		       + (-101./3. + 33.*ZETA2 + 351./2.*ZETA3 + _nf*(14./9.-2.*ZETA2)) * D(0,z)
		       ) );
  //double lrf = 2.*log(_muR_muF_ratio);
  //double muRF = _muR_muF_ratio;  _muR_muF_ratio = 1.;
  //Clog += (3*b1*FO_1_log(z,D)+4*b0*FO_2_log(z,D))*lrf + 6*b0*b0*FO_1_log(z,D)*lrf*lrf;
  //_muR_muF_ratio = muRF;
  return Clog;
}
dcomplex FO_3_log_APimprovement(dcomplex N, void *par) {
  return 1./6.*(pow(APimprovement(S1,N),3) - APimprovement(S1_cube,N))
    + APimprovement(S1,N)*APimprovement(S2,N) - APimprovement(S1_S2,N);
}
double FO_3_log_APimproved(double z, double D(int,double)) {
  double res = APimprovement(z)*FO_3_log(z,D);
  if(APimprExponent) res += InverseMellin(z,FO_3_log_APimprovement,NULL);
  return res;
}
dcomplex FO_3_Nsoftcoll_impr(dcomplex N, void *par) {
  return 2.*CA*log(N)/N*(S2(N)+S1(N)*S1(N)/2.);
}
double FO_3_Nsoftcoll_impr(double z) {
  return InverseMellin(z,FO_3_Nsoftcoll_impr,NULL);
}
double FO_3_delta_const(double mu) {
  double lf = -2.*log(_muF_mu_ratio);
  // not yet known in the full theory => use pointlike value.
  double W1 = _W1;  // (finite_mt ? _W1t : _W1 );
  double W2 = _W2;  // (finite_mt ? _W2t : _W2 );
  double W3 = _W3;  // (finite_mt ? _W3t : _W3 );
  //
  double c = W3 + W2*_cd1 + W1*_cd2 + _cd3;
  //c = (52.1348 - 1.97392*FO_1_delta_const()/M_PI) *M_PI*M_PI*M_PI;  // this corresponds to g0bar3 = 0 only for muF=mH
  if(__NNLL__) c = -FO_3_log(1, mell_const);
  //cout << c/M_PI/M_PI/M_PI << endl;
  //
  // factorization scale dependence
  c += (72.*ZETA3+18.*b0*ZETA2)*lf*lf*lf
    + (-4.*CA*A1*ZETA2-2.*b0*(B1-b1+6.*CA*CA*ZETA3)+b0*b0*(2.*CA*ZETA2+W1)-2.*CA*CA*W1*ZETA2+4.*CA*CA*CA*(ZETA2*ZETA2-6.*ZETA4))*lf*lf
    + (4.*CA*B1*ZETA2-6.*CA*b1*ZETA2-3.*W1*b1+2.*B2-2.*b2+2.*W1*B1+8.*CA*A1*ZETA3)*lf;
  // adding contributions from interference with C2
  c += lf*(2*CA*( 72. * 6*ZETA5
		  + (2.*_nf - 33.) * (-2*ZETA4)
		  + (67.+4*CA*W1 - 90.*ZETA2 - 10./3.*_nf) * ZETA3
		  + (-101./3. + 33.*ZETA2 + 351./2.*ZETA3 + _nf*(14./9.-2.*ZETA2)) * (-ZETA2)
		  )
	   -  2*b0* (_cd2 + W2 + _cd1*W1 ) );
  //double lrf = 2.*log(_muR_muF_ratio);
  //double muRF = _muR_muF_ratio;  _muR_muF_ratio = 1.;
  //c += (2*b2+3*b1*FO_1_delta_const()+4*b0*FO_2_delta_const(mu))*lrf + (7*b0*b1+6*b0*b0*FO_1_delta_const())*lrf*lrf + 4*b0*b0*b0*lrf*lrf*lrf;
  //_muR_muF_ratio = muRF;
  return c;
}
double FO_3_nonlog(double z, double mu) {
  double nonlog = 0.;  // unknown...
  return nonlog;
}
double FO_3_hard(double z, double mu) {
  //double hard = SF(z, mu, FO_3_sx_NLL_RC); // ok checked
  //double lrf = 2.*log(_muR_muF_ratio);
  //hard += FO_1_hard(z,mu) * (3*b1*lrf + 6*b0*b0*lrf*lrf);
  //
  double lf = -2.*log(_muF_mu_ratio);
  double c10 = sxc->c10(mu,mtop);
  double c20 = sxc->c20(mu,mtop);
  double c11 = sxc->c11(mu,mtop);
  double c30 = sxc->c30(mu,mtop);
  double c21 = sxc->c21(mu,mtop);
  if(MSbar) c30 += 8./3.*ZETA3;
  double c1 = 2*c10;
  double c2 = 2.*c20+c11;
  double c3 = 2.*(c30+c21);
  // scale dependence
  c30 += c20*lf + c1*lf*lf/4. + lf*lf*lf/6.;
  c21 += (c2-c20)*lf + 3.*c1*lf*lf/4. + lf*lf*lf/2.;
  c20 += (c1+lf)*lf/2.;
  c2  += (c1+lf)*lf*2.;
  c1  += 2*lf;
  //
  double e01 = CA;
  double e00 = (-11.*CA+2.*_nf*(2.*CF/CA-1.))/12.;
  double e11 = _nf*(26.*CF-23.*CA)/36.;
  double e22 = CA*( (18.*ZETA2-71.)*(CA-2*CF)*_nf + CA*CA*(54*ZETA3+99.*ZETA2-395) ) /108.;
  //
  double hard = (c3*e01*(e01*e01*pow(log(z),2)/2.-3*e01*e00*log(z)+3*e00*e00)
		 - (3*c30+c21)*2*b0*(2*e00*e01-e01*e01*log(z))
		 + 4*c30*b0*b0*e01
		 + 2*c2*(-e01*log(z)+e00)*e11
		 - 4*c20*b0*e11
		 - c1*e22*log(z) )/z;
  //
  return (hard - 12.*MomConsCoeff[2]) * pow(1-z, 2);
}




//
// PAY ATTENTION!!! they are computed assuming AP improvement at the exponent!!
double UncertaintyCoeff[3] = {0,0,0};
void ComputeUncertaintyCoeff(double mu) {
  UncertaintyCoeff[0] = FO_1_delta_const()   + real(APimprovement(S1,2));
  UncertaintyCoeff[1] = FO_2_delta_const(mu) + real(APimprovement(FO_2_log,2)
						    + 0.5*(pow(APimprovement(S1,2),2) - APimprovement(S1_sq,2)));
  UncertaintyCoeff[2] = FO_3_delta_const(mu) + real(APimprovement(FO_3_log,2)
						    + 1./6.*(pow(APimprovement(S1,2),3) - APimprovement(S1_cube,2))
						    + (APimprovement(S1,2)*APimprovement(S2,2) - APimprovement(S1_S2,2)));
}









double FO_log(double z, double D(int,double)) {
  double res = 0;
  if(order == 1) res = FO_1_log(z,D);
  if(order == 2) res = FO_2_log(z,D);
  if(order == 3) res = FO_3_log(z,D);
  return res;
}
double FO_log_APimproved(double z, double D(int,double)) {
  double res = 0;
  if(order == 1) res = FO_1_log_APimproved(z,D);
  if(order == 2) res = FO_2_log_APimproved(z,D);
  if(order == 3) res = FO_3_log_APimproved(z,D);
  return res;
}
double FO_Nsoftcoll_impr(double z) {
  double res = 0;
  if(order == 1) res = FO_1_Nsoftcoll_impr(z);
  if(order == 2) res = FO_2_Nsoftcoll_impr(z);
  if(order == 3) res = FO_3_Nsoftcoll_impr(z);
  return res;
}
double FO_nonlog(double z, double mu) {
  double res = 0;
  if(order == 1) res = FO_1_nonlog(z,mu);
  if(order == 2) res = FO_2_nonlog(z,mu);
  if(order == 3) res = FO_3_nonlog(z,mu);
  return res;
}
double FO_hard(double z, double mu) {
  if(!finite_mt) return 0;
  double res = 0;
  if(order == 1) res = FO_1_hard(z,mu);
  if(order == 2) res = FO_2_hard(z,mu);
  if(order == 3) res = FO_3_hard(z,mu);
  return res;
}
double FO_delta_const(double mu) {
  double res = 0;
  if(order == 1) res = FO_1_delta_const();
  if(order == 2) res = FO_2_delta_const(mu);
  if(order == 3) res = FO_3_delta_const(mu);
  return res;
}
double FO_qg(double z, double mu) {
  double res = 0;
  if(order == 1) res = FO_1_qg(z, mu);
  if(order == 2) res = FO_2_qg(z, mu);
  return res;
}
double FO_qqb(double z, double mu) {
  double res = 0;
  if(order == 1) res = FO_1_qqb(z, mu);
  if(order == 2) res = FO_2_qqb(z, mu);
  return res;
}
double FO_qq(double z, double mu) {
  double res = 0;
  if(order == 2) res = FO_2_qq(z, mu);
  return res;
}
double FO_qqp(double z, double mu) {
  double res = 0;
  if(order == 2) res = FO_2_qqp(z, mu);
  return res;
}








/////////////////////////////
//  LUMINOSITY
double lumi_gg(double x1, double x2, double muF) {
  return LHAPDF::xfx(NSET,x1,muF,0) * LHAPDF::xfx(NSET,x2,muF,0) /x1/x2;
}
double lumi_qg(double x1, double x2, double muF) {
  double xfx1[13], xfx2[13];
  LHAPDF::xfx(NSET, x1, muF, xfx1);
  LHAPDF::xfx(NSET, x2, muF, xfx2);
  double res = 0;
  for(int i=1; i<=5; i++) {
    res += (xfx1[6+i] + xfx1[6-i]) * xfx2[6+0];
    res += (xfx2[6+i] + xfx2[6-i]) * xfx1[6+0];
  }
  return res/x1/x2;
}
double lumi_qqb(double x1, double x2, double muF) {
  double xfx1[13], xfx2[13];
  LHAPDF::xfx(NSET, x1, muF, xfx1);
  LHAPDF::xfx(NSET, x2, muF, xfx2);
  double res = 0;
  for(int i=-5; i<=5; i++) {
    if(i==0) continue;
    res += xfx1[6+i] * xfx2[6-i];
  }
  return res/x1/x2;
}
double lumi_qq(double x1, double x2, double muF) {
  double xfx1[13], xfx2[13];
  LHAPDF::xfx(NSET, x1, muF, xfx1);
  LHAPDF::xfx(NSET, x2, muF, xfx2);
  double res = 0;
  for(int i=-5; i<=5; i++) {
    if(i==0) continue;
    res += xfx1[6+i] * xfx2[6+i];
  }
  return res/x1/x2;
}
double lumi_qqp(double x1, double x2, double muF) {
  double xfx1[13], xfx2[13];
  LHAPDF::xfx(NSET, x1, muF, xfx1);
  LHAPDF::xfx(NSET, x2, muF, xfx2);
  double res = 0;
  for(int i=-5; i<=5; i++) {
    for(int j=-5; j<=5; j++) {
      if(i!=j && i!=-j && i!=0 && j!=0) res += xfx1[6+i] * xfx2[6+j];
    }
  }
  return res/x1/x2;
}
/////////////////







////////////////////////////////////////
// integrating delta term
struct int1dp {
  double tau, mu;
};
double Lum_integrand(double x, void *par) {
  int1dp p = *(int1dp *) par;
  double tau = p.tau;
  double muF  = p.mu;
  double res = lumi_gg(tau/x,x,muF)/x;
  //cout << res << endl;
  return res;
}
double Lum(double tau, double muF) {
  int1dp p = { tau, muF };
  return gauss(Lum_integrand, tau, 1., INTEGRAL_PRECISION, NULL, &p);
}
////////////////////////////////////////




double distr(double z, double mu) {
  double res = 0;
  if(__full__)
    res = FO_log(z,Dk);
  else if(__soft_part__) {
    if(__log_type__ == 2)
      res = FO_log(z,DMPk);
    else
      res = FO_log(z,Dk);
  }
  //
  //double lrf = 2.*log(_muR_muF_ratio);
  //if(order == 2) res += 3*b0*FO_1_log(z,Dk)*lrf;
  //if(order == 3) res += (3*b1+6*b0*b0*lrf)*lrf*FO_1_log(z,Dk) + 4*b0*FO_2_log(z,Dk)*lrf;
  return res;
}
double reg(double z, double mu) {
  double res = 0;
  if(__full__) {
    res = FO_nonlog(z,mu);
  }
  else {
    if(__soft_part__ && __log_type__==1 && APtype!=0)
      res = FO_log_APimproved(z,Dtk) - FO_log(z,Dk);
    if(__soft_part__ && __log_type__==2 && __Nsoftcoll__)
      res = FO_Nsoftcoll_impr(z);
    if(__hard_part__)
      res += FO_hard(z,mu);
  }
  /*
  if(__full__ || __soft_part__) {
    double lrf = 2.*log(_muR_muF_ratio);
    if(order == 2) res += 3*b0*FO_1_nonlog(z,mu)*lrf;
    if(order == 3) res += (3*b1+6*b0*b0*lrf)*lrf*FO_1_nonlog(z,mu) + 4*b0*FO_2_nonlog(z,mu)*lrf;
  }
  */
  return res;
}
double delta(double mu) {
  double res = 0;
  if(__full__ || __soft_part__) {
    res = FO_delta_const(mu);
    if(__log_type__ == 2) res += FO_log(0, mell_const);
    /*
    double lrf = 2.*log(_muR_muF_ratio);
    if(order == 1) res += 2*b0*lrf;
    if(order == 2) res += (2*b1+3*b0*FO_1_delta_const())*lrf + 3*b0*b0*lrf*lrf;
    if(order == 3) res += (2*b2+3*b1*FO_1_delta_const()+4*b0*FO_2_delta_const(mu))*lrf
		     + (7*b0*b1+6*b0*b0*FO_1_delta_const())*lrf*lrf + 4*b0*b0*b0*lrf*lrf*lrf;
    */
  }
  return res;
}







// integrand of regular part
double reg_int(double z, double x, double tau, double mu) {
  return lumi_gg(tau/z/x, x, mu*_muF_mu_ratio) * reg(z, mu) /z /x;
}
double reg_int_chvar(double t1, double t2, double tau, double mu) {
  double z = 1.-(1.-tau)*t1;
  double x = 1.-(1.-tau/z)*t2;
  return reg_int(z, x, tau, mu) * (1-tau) * (1-tau/z);
}



// integrand of distributional part
double dlum_int(double z, double x, double tau, double mu) {
  return lumi_gg(tau/z/x, x, mu*_muF_mu_ratio) /z /x;
}
double dlum_int_chvar(double z, double t2, double tau, double mu) {
  double x = 1.-(1.-tau/z)*t2;
  return dlum_int(z, x, tau, mu) * (1-tau/z);
}
double distr_int(double z, double t2, double tau, double mu) {
  return ( dlum_int_chvar(z, t2, tau, mu) - dlum_int_chvar(1, t2, tau, mu) ) * distr(z, mu);
}
double distr_int_chvar(double t1, double t2, double tau, double mu) {
  double z = 1.-(1.-tau)*t1;
  return distr_int(z, t2, tau, mu) * (1-tau);
}

// integrand of the residual distributional part
double distr_residual_integrand(double z, void *par) {
  double mu  = *(double*)par;
  double res = distr(z,mu);
  return res;
}
double distr_residual(double tau, double mu) {
  double res = gauss(distr_residual_integrand, 0., tau, INTEGRAL_PRECISION, NULL, &mu);
  //cout << "!!! >>>>>" << res << endl;
  return res;
}





// non-gg channels contributions
double qg_int(double z, double x, double tau, double mu) {
  return lumi_qg(tau/z/x, x, mu*_muF_mu_ratio) * FO_qg(z, mu) /z /x;
}
double qg_int_chvar(double t1, double t2, double tau, double mu) {
  double z = 1.-(1.-tau)*t1;
  double x = 1.-(1.-tau/z)*t2;
  return qg_int(z, x, tau, mu) * (1-tau) * (1-tau/z);
}
//
double qqb_int(double z, double x, double tau, double mu) {
  return lumi_qqb(tau/z/x, x, mu*_muF_mu_ratio) * FO_qqb(z, mu) /z /x;
}
double qqb_int_chvar(double t1, double t2, double tau, double mu) {
  double z = 1.-(1.-tau)*t1;
  double x = 1.-(1.-tau/z)*t2;
  return qqb_int(z, x, tau, mu) * (1-tau) * (1-tau/z);
}
//
double qq_int(double z, double x, double tau, double mu) {
  return lumi_qq(tau/z/x, x, mu*_muF_mu_ratio) * FO_qq(z, mu) /z /x;
}
double qq_int_chvar(double t1, double t2, double tau, double mu) {
  double z = 1.-(1.-tau)*t1;
  double x = 1.-(1.-tau/z)*t2;
  return qq_int(z, x, tau, mu) * (1-tau) * (1-tau/z);
}
//
double qqp_int(double z, double x, double tau, double mu) {
  return lumi_qqp(tau/z/x, x, mu*_muF_mu_ratio) * FO_qqp(z, mu) /z /x;
}
double qqp_int_chvar(double t1, double t2, double tau, double mu) {
  double z = 1.-(1.-tau)*t1;
  double x = 1.-(1.-tau/z)*t2;
  return qqp_int(z, x, tau, mu) * (1-tau) * (1-tau/z);
}





bool __all_channels__ = false;
int integrandCUBA(const int *ndim, const double x[], const int *ncomp, double res[], void *pars) {
  int1dp p = *(int1dp *) pars;
  double tau = p.tau;
  double mu  = p.mu;
  res[0] = 0;
  res[0] += reg_int_chvar(x[0],x[1],tau,mu);
  if(__full__ || __soft_part__)
    res[0] += distr_int_chvar(x[0],x[1],tau,mu);
  if(__all_channels__ && __full__) {
    res[0] += qg_int_chvar(x[0],x[1],tau,mu);
    res[0] += qqb_int_chvar(x[0],x[1],tau,mu);
    if(order > 1)
      res[0] += qq_int_chvar(x[0],x[1],tau,mu) + qqp_int_chvar(x[0],x[1],tau,mu);
  }
  return 0;
}
double integral2d(double tau, double mu, double *err = NULL) {
  double res = 0;
  int1dp p = { tau, mu };
  //
  double cres[1], error[1], prob[1];
  double epsrel=INTEGRAL_PRECISION, epsabs=1.e-10;
  int last = 4;
  int verbose = 0;
  int nregions, neval, fail;
  //system("export CUBACORES=8");
  Cuhre(2, 1, integrandCUBA, &p,
	epsrel, epsabs, verbose | last,
	0, 500000, 9,
	&nregions, &neval, &fail, cres, error, prob);
  // Suave(2, 1, integrandCUBA, &p,
  // 	epsrel, epsabs, verbose | last, 0,
  // 	0, 500000, 1000, 25.,
  // 	&nregions, &neval, &fail, cres, error, prob);
  // Vegas(2, 1, integrandCUBA, &p,
  // 	epsrel, epsabs, verbose | last, 0,
  // 	0, 500000, 1000, 500, 1000,
  // 	0, NULL,
  // 	&neval, &fail, cres, error, prob);
  //  
  if(fail == -1) {
    cout << "  ERROR: dimension out of range." << endl;
  } else if(fail>0) {
    cout << "  ERROR: precision not reached." << endl;
  }
  double lum_tau = Lum(tau, mu*_muF_mu_ratio);
  res += cres[0]/lum_tau;
  if(err != NULL) *err = error[0]/lum_tau;
  return res;
}


double xs(double M, double sqrts, double *err = NULL) {
  double tau = M*M/sqrts/sqrts;
  double res = integral2d(tau, M, err);
  res += delta(M);
  res -= distr_residual(tau, M);
  return res;
}




dcomplex brn_q_ep0(dcomplex x) {
  dcomplex logx = log(x+I*1e-60);
  dcomplex HPL2 = 0.5*logx*logx;
  return -16.*(1.+x)*(1.+x)/x*HPL2 + 32.*(x-1.)*(x-1.)/x;
}
dcomplex sborn_mass(double mH) {
  int number_of_quarks = 1;
  vector<double> m;
  m.push_back(mtop);
  if(mbot>0) { number_of_quarks++; m.push_back(mbot); }
  if(mcha>0) { number_of_quarks++; m.push_back(mcha); }
  dcomplex res = 0;
  for(int i=0; i<number_of_quarks; i++) {
    dcomplex Wq = 4.*m[i]*m[i]/mH/mH;
    dcomplex Xq = -Wq/pow(1.+sqrt(1.-Wq),2);
    res += brn_q_ep0(Xq)*Wq*Wq /64.*(-3./4.);
  }
  return res/M_PI;
}
double xs_LO_prefactor(double mH) {
  double pref = 35.0309; // = Gf*pi/sqrt(2)/288 with the Gf in pb
  dcomplex born = sborn_mass(mH);
  return pref*real(born*conj(born));
}









void compute(double mH, double sqrts, double LO, double NLO, double NNLO,
	     double *rexact, double *rsoft2, double *rsoft1, double *rNsoft,
	     double *rhighenergy, double *momconsunc,
	     double *rsoft0, double *rsoftAP, double *rNsoftcoll, double *rNNLL) {
  double as = alphas(mH*_muF_mu_ratio*_muR_muF_ratio);
  //cout << "alpha_s = "<< as << endl;
  double fact = pow(as/M_PI,order) * LO;
  //
  string channel = (__all_channels__ ? "all channels:" : "gg channel:");
  if(!quiet) cout << "sigma = sigma0 * [ 1 + as sig1 + as^2 sig2 + as^3 sig3 + ...]" << endl
		  << channel << endl;
  string sassig[4] = { "  sig0 = ",
		       "  as sig1 = ",
		       "  as^2 sig2 = ",
		       "  as^3 sig3 = " };
  double err;
  double lrf = 2.*log(_muR_muF_ratio);
  double mur_dep_term = 0; // using terms of previous orders *at* mur
  NNLO -= NLO;
  NLO -= LO;
  double api = as/M_PI;
  if(order == 1) mur_dep_term = api*LO*2*b0*lrf;
  if(order == 2) mur_dep_term = api*(api*2*b1*LO + 3*b0*NLO) *lrf - api*api*3*b0*b0*LO*lrf*lrf;
  if(order == 3) mur_dep_term = api*(api*api*2*b2*LO + api*3*b1*NLO + 4*b0*NNLO) *lrf
		   - api*api*(api*7*b0*b1*LO + 6*b0*b0*NLO) *lrf*lrf + api*api*api*4*b0*b0*b0*LO*lrf*lrf*lrf;
  //cout << mur_dep_term << endl;
  //
  __full__ = true; // compute the full gg contribution to the hadronic xs
  __log_type__ = 0;
  APtype = 0;
  if(rexact != NULL && order < 3) {
    *rexact = xs(mH, sqrts, &err) * fact + mur_dep_term;
    err *= fact;
    if(!quiet) cout << sassig[order] << *rexact/LO << " +/- " << err/LO
		    << "  (full)" << endl;
  }
  //
  __full__ = false; // compute only approximations
  __soft_part__ = true; // turn on soft
  __hard_part__ = false; // leave off hard
  __log_type__ = 0;
  if(rsoft0 != NULL) {
    *rsoft0 = xs(mH, sqrts, &err) * fact + mur_dep_term;
    err *= fact;
    if(!quiet) cout << sassig[order] << *rsoft0/LO << " +/- " << err/LO
		    << "  (soft-0 only)" << endl;
  }
  //
  __log_type__ = 1;
  APtype = 2;
  if(rsoft2 != NULL) {
    *rsoft2 = xs(mH, sqrts, &err) * fact + mur_dep_term;
    err *= fact;
    if(!quiet) cout << sassig[order] << *rsoft2/LO << " +/- " << err/LO
		    << "  (soft_2 only)" << endl;
  }
  //
  __log_type__ = 1;
  APtype = 1;
  if(rsoft1 != NULL) {
    *rsoft1 = xs(mH, sqrts, &err) * fact + mur_dep_term;
    err *= fact;
    if(!quiet) cout << sassig[order] << *rsoft1/LO << " +/- " << err/LO
		    << "  (soft_1 only)" << endl;
  }
  //
  __log_type__ = 1;
  APtype = -1;
  if(rsoftAP != NULL) {
    *rsoftAP = xs(mH, sqrts, &err) * fact + mur_dep_term;
    err *= fact;
    if(!quiet) cout << sassig[order] << *rsoftAP/LO << " +/- " << err/LO
		    << "  (soft_inf (full AP) only)" << endl;
  }
  //
  __log_type__ = 2;
  APtype = 0;
  if(rNsoft != NULL) {
    *rNsoft = xs(mH, sqrts, &err) * fact + mur_dep_term;
    err *= fact;
    if(!quiet) cout << sassig[order] << *rNsoft/LO << " +/- " << err/LO
		    << "  (N-soft only)" << endl;
  }
  //
  if(rNsoftcoll != NULL) {
    __Nsoftcoll__ = true;
    *rNsoftcoll = xs(mH, sqrts, &err) * fact + mur_dep_term;
    err *= fact;
    if(!quiet) cout << sassig[order] << *rNsoftcoll/LO << " +/- " << err/LO
		    << "  (N-soft-collinear only)" << endl;
    __Nsoftcoll__ = false;
  }
  //
  if(rNNLL != NULL) {
    __NNLL__ = true;
    *rNNLL = xs(mH, sqrts, &err) * fact + mur_dep_term;
    err *= fact;
    if(!quiet) cout << sassig[order] << *rNNLL/LO << " +/- " << err/LO
		    << "  (N-soft-collinear only)" << endl;
    __NNLL__ = false;
  }
  //
  __soft_part__ = false; // turn soft off
  __hard_part__ = true;  // turn hard on
  if(rhighenergy != NULL) {
    // momentum conservation
    ComputeMomConsCoeff(mH);
    *rhighenergy = xs(mH, sqrts, &err) * fact;
    err *= fact;
    if(!quiet) cout << sassig[order] << *rhighenergy/LO << " +/- " << err/LO
		    << "  (high-energy --- must be added to a soft)" << endl;
  }
  //
  if(momconsunc != NULL) {
    if(rhighenergy != NULL) {
      *momconsunc = *rhighenergy;
    } else {
      ComputeMomConsCoeff(mH);
      *momconsunc = xs(mH, sqrts, &err) * fact;
    }
    double MCcoeff = MomConsCoeff[order-1];
    ResetMomConsCoeff();
    *momconsunc -= xs(mH, sqrts, &err) * fact; // mom-cons function
    //
    APtype = 2;  __log_type__ = 1;
    ComputeUncertaintyCoeff(mH);
    //cout << UncertaintyCoeff[order-1]/MCcoeff << endl;
    //
    *momconsunc *= -UncertaintyCoeff[order-1]/MCcoeff;
    //cout << *momconsunc << endl;
    err *= fact*UncertaintyCoeff[order-1]/MCcoeff;
    if(!quiet) cout << sassig[order] << *momconsunc*0.05/LO << " +/- " << err*0.05/LO
		    << "  (momcons uncertainty)" << endl;
  }
}














// interface to LHAPDF
int pdf_init(string SET_NAME, int SUBSET, LHAPDF::Verbosity v) {
  setVerbosity(v);
  LHAPDF::initPDFSet(NSET, SET_NAME, SUBSET);
  //LHAPDF::usePDFMember(SUBSET);
  cout << ">>> Using PDFs:  " << SET_NAME << " ,  subset: " << SUBSET
       << " ,  with alphs_s(mZ) = " << LHAPDF::alphasPDF(mZ) << endl;
  return 0;
}
//




